#include<Servo.h>
#include <AFMotor.h>

//Create an object to control each motor
AF_DCMotor rightBack(1);                          
AF_DCMotor rightFront(2);
AF_DCMotor leftFront(3);
AF_DCMotor leftBack(4);

Servo servoBox;
Servo servoLook;

//detecting distance
int distance = 0;
int position = 0;
byte trig = A2;
byte echo = A1;
byte maxDist = 150;
byte stopDist = 50;
float timeOut = 2*(maxDist+10)/100/340*1000000;   //Maximum time to wait for a return signal

//set motor speed
byte motorSpeed = 200; 
int motorOffset = 10; 
int turnSpeed = 40;   

//knocking machenism
const int piezo = A3;
const int quietknock = 8;
const int loudknock = 100;
int knockval;
bool locked = true;
int numberOfknocks = 0;
const int greenled = 13;
const int redled = 2;

void setup() {
  Serial.begin(9600);
  servoBox.attach(10);
  servoLook.attach(9);
  pinMode(trig,OUTPUT); 
  pinMode(echo,INPUT);
  pinMode(redled, OUTPUT);
  pinMode(greenled, OUTPUT);
  digitalWrite(greenled, HIGH);
  
  //Set the motors to the motor speed
  rightBack.setSpeed(motorSpeed);                 
  rightFront.setSpeed(motorSpeed);
  leftFront.setSpeed(motorSpeed);
  leftBack.setSpeed(motorSpeed);

  //Ensure all motors are stopped
  rightBack.run(RELEASE);                         
  rightFront.run(RELEASE);
  leftFront.run(RELEASE);
  leftBack.run(RELEASE);

  servoBox.write(0);

}

void loop() {
  //if the box is closed
  if (locked==true){
    //Check the left and right object distances and get the turning instruction
    int turnDir = checkDirection();                 
    Serial.print(turnDir);

    //Turn left, turn around or turn right depending on the instruction
    switch (turnDir)                                
    {
      //Turn left
      case 0:
        turnLeft (400);
        break;
      //Turn around
      case 1:
        turnLeft (700);
         break;
      //Turn right
      case 2:
        turnRight (400);
        break;
    } 
    
    distance = getDistance();
    Serial.print("distance");                   
    Serial.println(distance);    

    position = 0;
    servoBox.write(position);
    
    //If there are no objects within the stopping distance, move forward
    if(distance >= stopDist)                        
    {
      moveForward();
    }
    
    //Keep checking the object distance until it is within the minimum stopping distance
    while(distance >= stopDist)                     
    {
      distance = getDistance();
      delay(50);
    }

    //Stop, turn around, and open the box
    stopMove();
    delay(1000);
    moveBackward();
    turnLeft(2800);
    stopMove(); 
    position = 90;
    servoBox.write(position);
    locked = false;
    delay(50);
  }

  //if the box is open
  if (locked == false){
    knockval = analogRead(piezo);
    
    //if the amount of knocks is less than three times, keep checking if there are valid knocks
    if(numberOfknocks < 3 && knockval > 0){
      position = 90;
      servoBox.write(position);
      digitalWrite(greenled, LOW);
      digitalWrite(redled,HIGH);
      if(checkForKnock(knockval) == true){
        numberOfknocks++;
      } 
      Serial.print(numberOfknocks);
      Serial.println(" more knocks to go ");
    }

    //if the amount of knocks is more than three times, close the box
    if (numberOfknocks >= 3){
      locked = true;
      position = 0;
      servoBox.write(position);
      delay(20);
      digitalWrite(greenled, HIGH);
      digitalWrite(redled,LOW);
      Serial.println(" the box is unlocked! ");
      numberOfknocks = 0;
      delay(1000);
    }
  }
}

//Measure the distance to an object
int getDistance()                                   
{
  unsigned long pulseTime;                          //Create a variable to store the pulse travel time
  int distance;                                     //Create a variable to store the calculated distance
  digitalWrite(trig, HIGH);                         //Generate a 10 microsecond pulse
  delayMicroseconds(10);
  digitalWrite(trig, LOW);
  pulseTime = pulseIn(echo, HIGH, timeOut);         //Measure the time for the pulse to return
  distance = (float)pulseTime * 340 / 2 / 10000;    //Calculate the object distance based on the pulse time
  return distance;
}

//Set all motors to run forward
void moveForward()                                
{
  //Set the motors to the motor speed
  rightBack.setSpeed(motorSpeed);                 
  rightFront.setSpeed(motorSpeed);
  leftFront.setSpeed(motorSpeed);
  leftBack.setSpeed(motorSpeed);
  rightBack.run(BACKWARD);
  rightFront.run(BACKWARD);
  leftFront.run(BACKWARD);
  leftBack.run(BACKWARD);
}

void moveBackward()
{
  rightBack.run(FORWARD);
  rightFront.run(FORWARD);
  leftFront.run(FORWARD);
  leftBack.run(FORWARD);
  delay(250);
}

//Set all motors to stop
void stopMove()                                   
{
  rightBack.run(RELEASE);
  rightFront.run(RELEASE);
  leftFront.run(RELEASE);
  leftBack.run(RELEASE);
}

//Set motors to turn left for the specified duration then stop
void turnLeft(int duration)                                 
{
  //Set the motors to the motor speed
  rightBack.setSpeed(motorSpeed+turnSpeed);                 
  rightFront.setSpeed(motorSpeed+turnSpeed);
  leftFront.setSpeed(motorSpeed+motorOffset+turnSpeed);
  leftBack.setSpeed(motorSpeed+motorOffset+turnSpeed);
  rightBack.run(BACKWARD);
  rightFront.run(BACKWARD);
  leftFront.run(FORWARD);
  leftBack.run(FORWARD);
  delay(duration);
}

//Set motors to turn right for the specified duration then stop
void turnRight(int duration)                                
{
  //Set the motors to the motor speed
  rightBack.setSpeed(motorSpeed+turnSpeed);                 
  rightFront.setSpeed(motorSpeed+turnSpeed);
  leftFront.setSpeed(motorSpeed+motorOffset+turnSpeed);
  leftBack.setSpeed(motorSpeed+motorOffset+turnSpeed);
  rightBack.run(FORWARD);
  rightFront.run(FORWARD);
  leftFront.run(BACKWARD);
  leftBack.run(BACKWARD);
  delay(duration);
}

//Check the left and right directions and decide which way to turn
int checkDirection()                                            
{
  int distances [2] = {0,0};                                    //Left and right distances
  int turnDir = 1;                                              //Direction to turn, 0 left, 1 reverse, 2 right
  servoLook.write(0);                                         //Turn servo to look left
  delay(500);
  distances [0] = getDistance();                                //Get the left object distance
  servoLook.write(180);                                           //Turn servo to look right
  delay(500);
  servoLook.write(90); 
  delay(500);
  distances [1] = getDistance();                                //Get the right object distance
  if (distances[0]>=200 && distances[1]>=200)                   //If both directions are clear, turn left
    turnDir = 0;
  else if (distances[0]<=stopDist && distances[1]<=stopDist)    //If both directions are blocked, turn around
    turnDir = 1;
  else if (distances[0]>=distances[1])                          //If left has more space, turn left
    turnDir = 0;
  else if (distances[0]<distances[1])                           //If right has more space, turn right
    turnDir = 2;
  return turnDir;
}

//check if the knock is valid
boolean checkForKnock(int value){
  if(value > quietknock && value < loudknock){
    delay(50);
    Serial.print("valid knock of value");
    Serial.println(value);
    delay(250);
    return true;
  }
  else{
   Serial.print(" Bad knock value ");
   Serial.println(value);  
   return false;
  }
}
